# pyright: reportOptionalMemberAccess=false
# mypy: disable-error-code=no-untyped-call
import base64
import hashlib
import http.client
import ipaddress
import io
import logging
import socket
import zlib
from dataclasses import dataclass
from io import BytesIO
from math import ceil
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Iterable,
    Optional,
    TypeAlias,
    TypeGuard,
    Union,
    no_type_check,
)
from urllib.parse import urlsplit
from urllib.request import (
    HTTPHandler,
    HTTPRedirectHandler,
    HTTPSHandler,
    ProxyHandler,
    Request,
    build_opener,
)

from .enums import ResourceAccessPolicy
from .errors import FPDFException, FPDFResourceAccessError
from .image_datastructures import (
    ImageCache,
    ImageFilter,
    RasterImageInfo,
    VectorImageInfo,
)
from .svg import SVGObject, SVGLimits
from .util import ImageType

try:
    from PIL import Image, ImageCms, TiffImagePlugin, features as PIL_features
except ImportError:
    Image = None  # type: ignore[assignment]
    ImageCms = None  # type: ignore[assignment]
    TiffImagePlugin = None  # type: ignore[assignment]
    PIL_features = None  # type: ignore[assignment]

try:
    from PIL.Image import Resampling

    RESAMPLE = Resampling.LANCZOS
except (ImportError, AttributeError):
    RESAMPLE = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from PIL.Image import Image as PILImage
    from PIL.ImageCms import ImageCmsProfile
else:
    PILImage: TypeAlias = Any

try:
    import numpy
except (ImportError, RuntimeError):
    numpy = None  # type: ignore[assignment]


@dataclass
class ImageSettings:
    # Passed to zlib.compress() - In range 0-9 - Default is currently equivalent to 6:
    compression_level: int = -1
    # Applied to remote HTTP(S) image fetches. Set to None to use the Python default.
    network_timeout: float | None = 10.0


LOGGER = logging.getLogger(__name__)
SUPPORTED_IMAGE_FILTERS = ("AUTO", "FlateDecode", "DCTDecode", "JPXDecode", "LZWDecode")
SETTINGS = ImageSettings()


def _resource_scope_for_ip(
    address: ipaddress.IPv4Address | ipaddress.IPv6Address,
) -> tuple[ResourceAccessPolicy, str]:
    if address.is_global:
        return ResourceAccessPolicy.REMOTE_PUBLIC, "public"
    return ResourceAccessPolicy.REMOTE_PRIVATE, "private"


def _resolve_hostname(
    hostname: str,
) -> tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...]:
    try:
        addr_info = socket.getaddrinfo(hostname, None, type=socket.SOCK_STREAM)
    except socket.gaierror as error:
        raise FPDFResourceAccessError(
            f"Could not resolve remote resource hostname: {hostname!r}"
        ) from error
    addresses: list[ipaddress.IPv4Address | ipaddress.IPv6Address] = []
    seen: set[str] = set()
    for addr in addr_info:
        sockaddr = addr[4]
        address_value = sockaddr[0]
        if not isinstance(address_value, str):
            continue
        address = address_value.split("%", 1)[0]
        if address in seen:
            continue
        seen.add(address)
        addresses.append(ipaddress.ip_address(address))
    if not addresses:
        raise FPDFResourceAccessError(
            f"Could not resolve any IP address for remote resource hostname: {hostname!r}"
        )
    return tuple(addresses)


def _validate_remote_url_access(
    url: str,
    resource_access_policy: ResourceAccessPolicy,
) -> tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...]:
    parsed_url = urlsplit(url)
    if parsed_url.scheme not in ("http", "https") or not parsed_url.hostname:
        raise FPDFResourceAccessError(f"Unsupported remote resource URL: {url!r}")
    if (
        ResourceAccessPolicy.REMOTE_PUBLIC not in resource_access_policy
        and ResourceAccessPolicy.REMOTE_PRIVATE not in resource_access_policy
    ):
        raise FPDFResourceAccessError(
            f"Remote resource access is disabled by resource_access_policy: {url!r}"
        )

    addresses = _resolve_hostname(parsed_url.hostname)
    for address in addresses:
        required_policy, scope = _resource_scope_for_ip(address)
        if required_policy in resource_access_policy:
            continue
        raise FPDFResourceAccessError(
            "Remote resource access is blocked by resource_access_policy: "
            f"{url!r} resolves to {scope} address {address}"
        )
    return addresses


def _validate_resource_access(
    filename: Any,
    resource_access_policy: ResourceAccessPolicy,
) -> None:
    if isinstance(filename, Path):
        if ResourceAccessPolicy.LOCAL_FILES not in resource_access_policy:
            raise FPDFResourceAccessError(
                "Local file access is disabled by resource_access_policy: "
                f"{filename!r}"
            )
        return
    if not isinstance(filename, str):
        return
    if filename.startswith("data:"):
        return
    if filename.startswith(("http://", "https://")):
        _validate_remote_url_access(filename, resource_access_policy)
        return
    if ResourceAccessPolicy.LOCAL_FILES not in resource_access_policy:
        raise FPDFResourceAccessError(
            f"Local file access is disabled by resource_access_policy: {filename!r}"
        )


class _ResourceAccessPolicyRedirectHandler(HTTPRedirectHandler):
    def __init__(
        self,
        resource_access_policy: ResourceAccessPolicy,
        pinned_addresses_by_url: dict[
            str, tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...]
        ],
    ) -> None:
        super().__init__()
        self.resource_access_policy = resource_access_policy
        self.pinned_addresses_by_url = pinned_addresses_by_url

    def redirect_request(
        self,
        req: Any,
        fp: Any,
        code: Any,
        msg: Any,
        headers: Any,
        newurl: str,
    ) -> Optional[Request]:
        self.pinned_addresses_by_url[newurl] = _validate_remote_url_access(
            newurl, self.resource_access_policy
        )
        return super().redirect_request(req, fp, code, msg, headers, newurl)


class _PinnedRemoteHTTPConnection(http.client.HTTPConnection):
    def __init__(
        self,
        *args: Any,
        pinned_addresses: tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._pinned_addresses = pinned_addresses
        self._create_connection = self._create_pinned_connection

    def _create_pinned_connection(
        self,
        address: tuple[str, int],
        timeout: Any = None,
        source_address: tuple[str, int] | None = None,
    ) -> socket.socket:
        _hostname, port = address
        last_error: OSError | None = None
        for pinned_address in self._pinned_addresses:
            try:
                return socket.create_connection(
                    (str(pinned_address), port), timeout, source_address
                )
            except OSError as error:
                last_error = error
        if last_error is not None:
            raise last_error
        raise OSError("No pinned address available for remote resource")


class _PinnedRemoteHTTPSConnection(
    _PinnedRemoteHTTPConnection, http.client.HTTPSConnection
):
    pass


class _ResourceAccessPolicyHTTPHandler(HTTPHandler):
    def __init__(
        self,
        resource_access_policy: ResourceAccessPolicy,
        pinned_addresses_by_url: dict[
            str, tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...]
        ],
    ) -> None:
        super().__init__()
        self.resource_access_policy = resource_access_policy
        self.pinned_addresses_by_url = pinned_addresses_by_url

    def http_open(self, req: Request) -> Any:
        pinned_addresses = self.pinned_addresses_by_url.pop(
            req.get_full_url(), None
        ) or _validate_remote_url_access(
            req.get_full_url(), self.resource_access_policy
        )

        def pinned_connection(*args: Any, **kwargs: Any) -> _PinnedRemoteHTTPConnection:
            return _PinnedRemoteHTTPConnection(
                *args, pinned_addresses=pinned_addresses, **kwargs
            )

        return self.do_open(pinned_connection, req)


class _ResourceAccessPolicyHTTPSHandler(HTTPSHandler):
    def __init__(
        self,
        resource_access_policy: ResourceAccessPolicy,
        pinned_addresses_by_url: dict[
            str, tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...]
        ],
    ) -> None:
        super().__init__()
        self.resource_access_policy = resource_access_policy
        self.pinned_addresses_by_url = pinned_addresses_by_url

    def https_open(self, req: Request) -> Any:
        pinned_addresses = self.pinned_addresses_by_url.pop(
            req.get_full_url(), None
        ) or _validate_remote_url_access(
            req.get_full_url(), self.resource_access_policy
        )

        def pinned_connection(
            *args: Any, **kwargs: Any
        ) -> _PinnedRemoteHTTPSConnection:
            return _PinnedRemoteHTTPSConnection(
                *args, pinned_addresses=pinned_addresses, **kwargs
            )

        return self.do_open(pinned_connection, req)


def _build_remote_resource_opener(
    resource_access_policy: ResourceAccessPolicy,
    pinned_addresses_by_url: dict[
        str, tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...]
    ],
) -> Any:
    return build_opener(
        # Proxying would make the validated origin address ambiguous.
        ProxyHandler({}),
        _ResourceAccessPolicyRedirectHandler(
            resource_access_policy, pinned_addresses_by_url
        ),
        _ResourceAccessPolicyHTTPHandler(
            resource_access_policy, pinned_addresses_by_url
        ),
        _ResourceAccessPolicyHTTPSHandler(
            resource_access_policy, pinned_addresses_by_url
        ),
    )


# fmt: off
TIFFBitRevTable = [
    0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90,
    0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 0x08, 0x88, 0x48, 0xC8,
    0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8,
    0x78, 0xF8, 0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
    0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 0x0C, 0x8C,
    0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC,
    0x3C, 0xBC, 0x7C, 0xFC, 0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2,
    0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
    0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A,
    0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA, 0x06, 0x86, 0x46, 0xC6,
    0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6,
    0x76, 0xF6, 0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
    0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE, 0x01, 0x81,
    0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1,
    0x31, 0xB1, 0x71, 0xF1, 0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9,
    0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
    0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95,
    0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5, 0x0D, 0x8D, 0x4D, 0xCD,
    0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD,
    0x7D, 0xFD, 0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
    0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 0x0B, 0x8B,
    0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB,
    0x3B, 0xBB, 0x7B, 0xFB, 0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7,
    0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
    0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F,
    0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF,
]
# fmt: on

LZW_CLEAR_TABLE_MARKER = 256  # Special code to indicate table reset
LZW_EOD_MARKER = 257  # End-of-data marker
LZW_INITIAL_BITS_PER_CODE = 9  # Initial code bit width
LZW_MAX_BITS_PER_CODE = 12  # Maximum code bit width


def preload_image(
    image_cache: ImageCache,
    name: ImageType,
    dims: Optional[tuple[float, float]] = None,
    resource_access_policy: ResourceAccessPolicy = ResourceAccessPolicy.DEFAULT,
    svg_limits: Optional[SVGLimits] = None,
) -> tuple[
    str,
    Union[SVGObject, "PILImage", bytes, BinaryIO, Path, None],
    RasterImageInfo | VectorImageInfo,
]:
    """
    Read an image and load it into memory.

    For raster images: following this call, the image is inserted in `image_cache.images`,
    and following calls to `fpdf.fpdf.FPDF.image()` will re-use the same cached values, without re-reading the image.

    For vector images: the data is loaded and the metadata extracted.

    Args:
        image_cache: an `ImageCache` instance, usually the `.image_cache` attribute of a `FPDF` instance.
        name: either a string representing a file path to an image, an URL to an image,
            an io.BytesIO, or a instance of `PIL.Image.Image`.
        dims (tuple[int, int]): optional dimensions as a tuple (width, height) to resize the image
            (raster only) before storing it in the PDF.

    Returns: A tuple, consisting of 3 values: the name, the image data,
        and an instance of a subclass of `ImageInfo`.
    """
    if isinstance(name, Path):
        if ResourceAccessPolicy.LOCAL_FILES not in resource_access_policy:
            raise FPDFResourceAccessError(
                "Local file access is disabled by resource_access_policy: " f"{name!r}"
            )
        name = str(name)
        if not name.endswith(".svg"):
            info = image_cache.images.get(name)
            if info is not None:
                info["usages"] = info["usages"] + 1  # type: ignore[operator]
                return name, None, info
    else:
        _validate_resource_access(name, resource_access_policy)

    # Identify and load SVG data:
    if isinstance(name, str) and name.endswith(".svg"):
        try:
            return get_svg_info(
                name,
                load_image(name, resource_access_policy=resource_access_policy),
                image_cache=image_cache,
                resource_access_policy=resource_access_policy,
                svg_limits=svg_limits,
            )
        except FPDFResourceAccessError:
            raise
        except FPDFException:
            raise
        except Exception as error:
            raise ValueError(f"Could not parse file: {name}") from error
    if isinstance(name, bytes) and _is_svg(name.strip()):
        return get_svg_info(
            "vector_image",
            io.BytesIO(name),
            image_cache=image_cache,
            resource_access_policy=resource_access_policy,
            svg_limits=svg_limits,
        )
    if isinstance(name, io.BytesIO) and _is_svg(name.getvalue().strip()):
        return get_svg_info(
            "vector_image",
            name,
            image_cache=image_cache,
            resource_access_policy=resource_access_policy,
            svg_limits=svg_limits,
        )

    # Load raster data.
    img: Union["PILImage", bytes, BinaryIO, Path, None]
    raster_name: str
    if isinstance(name, str):
        raster_name, img = name, None
    elif _is_pil_image(name):
        bytes_ = name.tobytes()
        img_hash = hashlib.new("md5", usedforsecurity=False)  # nosec B324
        img_hash.update(bytes_)
        raster_name, img = img_hash.hexdigest(), name
    elif isinstance(name, (bytes, io.BytesIO)):
        bytes_ = name.getvalue() if isinstance(name, io.BytesIO) else name
        bytes_ = bytes_.strip()
        img_hash = hashlib.new("md5", usedforsecurity=False)  # nosec B324
        img_hash.update(bytes_)
        raster_name, img = img_hash.hexdigest(), name
    elif _is_binary_stream(name):
        raster_name, img = str(name), name
    else:
        raster_name, img = str(name), None
    info = image_cache.images.get(raster_name)
    if info is not None:
        info["usages"] = info["usages"] + 1  # type: ignore[operator]
    else:
        info = get_img_info(
            raster_name,
            img,
            image_cache.image_filter,
            dims,
            resource_access_policy=resource_access_policy,
        )
        info["i"] = len(image_cache.images) + 1
        info["usages"] = 1
        info["iccp_i"] = None
        iccp = info.get("iccp")
        if iccp is not None:
            LOGGER.debug(
                "ICC profile found for image %s - It will be inserted in the PDF document",
                raster_name,
            )
            if iccp in image_cache.icc_profiles:
                info["iccp_i"] = image_cache.icc_profiles[
                    iccp
                ]  # pyright: ignore[reportArgumentType]
            else:
                iccp_i = len(image_cache.icc_profiles)
                image_cache.icc_profiles[iccp] = iccp_i  # type: ignore[index]
                info["iccp_i"] = iccp_i
            info["iccp"] = None
        image_cache.images[raster_name] = info
    return raster_name, img, info


def _is_svg(bytes_: bytes) -> bool:
    return bytes_.startswith(b"<?xml ") or bytes_.startswith(b"<svg ")


def _is_pil_image(obj: Any) -> TypeGuard[PILImage]:
    return Image is not None and isinstance(obj, Image.Image)


def _is_binary_stream(obj: Any) -> TypeGuard[BinaryIO]:
    return hasattr(obj, "read") and not isinstance(obj, (str, Path))


def load_image(
    filename: str | Path | BinaryIO,
    resource_access_policy: ResourceAccessPolicy = ResourceAccessPolicy.DEFAULT,
) -> BinaryIO:
    """
    This method is used to load external resources, such as images.
    It is automatically called when resource added to document by `fpdf.fpdf.FPDF.image()`.
    It always return a BytesIO buffer.
    """
    # if a file-like object is passed in, use it directly or copy it into a BytesIO buffer
    if isinstance(filename, (BytesIO, io.BufferedIOBase, BinaryIO)):
        return filename
    if _is_binary_stream(filename):
        # Copy other file-like objects into a BytesIO so downstream code can seek/read freely
        return BytesIO(filename.read())
    if isinstance(filename, Path):
        filename = str(filename)
    # Resource access is governed by the active resource_access_policy.
    if filename.startswith(("http://", "https://")):
        pinned_addresses_by_url = {
            filename: _validate_remote_url_access(filename, resource_access_policy)
        }
        opener = _build_remote_resource_opener(
            resource_access_policy, pinned_addresses_by_url
        )
        # disabling bandit & semgrep rules as permitted schemes are whitelisted:
        # nosemgrep: python.lang.security.audit.dynamic-urllib-use-detected.dynamic-urllib-use-detected
        with opener.open(
            filename, timeout=SETTINGS.network_timeout
        ) as url_file:  # nosec B310
            return BytesIO(url_file.read())
    elif filename.startswith("data:"):
        return _decode_base64_image(filename)
    _validate_resource_access(filename, resource_access_policy)
    with open(filename, "rb") as local_file:
        return BytesIO(local_file.read())


def _decode_base64_image(base64Image: str) -> BytesIO:
    "Decode the base 64 image string into an io byte stream."
    frags = base64Image.split("base64,")
    if len(frags) != 2:
        raise NotImplementedError("Unsupported non-base64 image data")
    imageData = frags[1]
    decodedData = base64.b64decode(imageData)
    return BytesIO(decodedData)


@no_type_check
def is_iccp_valid(iccp: bytes, filename: str | Path) -> bool:
    "Checks the validity of an ICC profile"
    try:
        profile: "ImageCmsProfile" = ImageCms.getOpenProfile(BytesIO(iccp))
    except ImageCms.PyCMSError:
        LOGGER.info("Invalid ICC Profile in file %s", filename)
        return False
    color_space = str(profile.profile.xcolor_space).strip()
    if color_space not in ("GRAY", "RGB"):
        LOGGER.info(
            "Unsupported color space %s in ICC Profile of file %s - cf. issue #711",
            color_space,
            filename,
        )
        return False
    return True


def get_svg_info(
    filename: str,
    img: BinaryIO,
    image_cache: ImageCache,
    resource_access_policy: ResourceAccessPolicy = ResourceAccessPolicy.DEFAULT,
    svg_limits: Optional[SVGLimits] = None,
) -> tuple[str, SVGObject, VectorImageInfo]:
    img.seek(0)
    svg_data = img.read()
    svg = SVGObject(
        svg_data,
        image_cache=image_cache,
        resource_access_policy=resource_access_policy,
        svg_limits=svg_limits,
    )
    if svg.viewbox:
        _, _, w, h = svg.viewbox
    else:
        w = h = 0.0
    if svg.width:
        w = svg.width
    if svg.height:
        h = svg.height
    info = VectorImageInfo(data=svg, w=w, h=h)
    return filename, svg, info


def get_img_info(
    filename: Union[str, BinaryIO, Path],
    img: Union["PILImage", bytes, BinaryIO, Path, str, None] = None,
    image_filter: ImageFilter = "AUTO",
    dims: Optional[tuple[float, float]] = None,
    resource_access_policy: ResourceAccessPolicy = ResourceAccessPolicy.DEFAULT,
) -> RasterImageInfo:
    """
    Args:
        filename: in a format that can be passed to load_image
        img: optional `bytes`, `BytesIO` or `PIL.Image.Image` instance
        image_filter (str): one of the SUPPORTED_IMAGE_FILTERS
    """
    if Image is None:
        raise EnvironmentError("Pillow not available - fpdf2 cannot insert images")

    is_pil_img = True
    keep_bytes_io_open = False
    # Flag to check whether a cmyk image is jpeg or not, if set to True the decode array
    # is inverted in output.py
    jpeg_inverted = False
    img_raw_data: Optional[BinaryIO] = None
    if not img or isinstance(img, (Path, str)):
        img_raw_data = load_image(
            filename, resource_access_policy=resource_access_policy
        )
        img = Image.open(img_raw_data)
        is_pil_img = False
    elif not _is_pil_image(img):
        keep_bytes_io_open = isinstance(img, BytesIO)
        if isinstance(img, bytes):
            img_raw_data = BytesIO(img)
        else:
            img_raw_data = img  # type: ignore[assignment]
        assert img_raw_data is not None
        img = Image.open(img_raw_data)
        is_pil_img = False
    assert _is_pil_image(img)

    img_altered = False
    if dims:
        img = img.resize(dims, resample=RESAMPLE)  # type: ignore[arg-type]
        img_altered = True

    if image_filter == "AUTO":
        # Very simple logic for now:
        if img.format == "JPEG":
            image_filter = "DCTDecode"
        elif img.mode == "1" and PIL_features.check("libtiff"):
            # The 2nd condition prevents from running in a bug sometimes,
            # cf. test_transcode_monochrome_and_libtiff_support_custom_tags()
            image_filter = "CCITTFaxDecode"
        else:
            image_filter = "FlateDecode"

    if img.mode in ("P", "PA") and image_filter != "FlateDecode":
        img = img.convert("RGBA")

    if img.mode not in ("1", "L", "LA", "RGB", "RGBA", "P", "PA", "CMYK"):
        img = img.convert("RGBA")
        img_altered = True

    if img.mode in ("P", "RGBA") and image_filter == "LZWDecode":
        img = img.convert("RGB")
    elif img.mode in ("LA") and image_filter == "LZWDecode":
        img = img.convert("L")

    w, h = img.size
    info = RasterImageInfo()

    iccp = None
    if "icc_profile" in img.info:
        if is_iccp_valid(img.info["icc_profile"], filename):
            iccp = img.info["icc_profile"]

    if img_raw_data is not None and not img_altered:
        # if we can use the original image bytes directly we do (JPEG and group4 TIFF only):
        if img.format == "JPEG" and image_filter == "DCTDecode":
            if img.mode in ("RGB", "RGBA"):
                dpn, bpc, colspace = 3, 8, "DeviceRGB"
            elif img.mode == "CMYK":
                dpn, bpc, colspace = 4, 8, "DeviceCMYK"
                jpeg_inverted = True
            elif img.mode == "L":
                dpn, bpc, colspace = 1, 8, "DeviceGray"
            else:
                raise ValueError(f"Unsupported image mode: {img.mode}")
            img_raw_data.seek(0)
            info.update(
                {
                    "data": img_raw_data.read(),
                    "w": w,
                    "h": h,
                    "cs": colspace,
                    "iccp": iccp,
                    "dpn": dpn,
                    "bpc": bpc,
                    "f": image_filter,
                    "inverted": jpeg_inverted,
                    "dp": f"/Predictor 15 /Colors {dpn} /Columns {w}",
                }
            )
            return info
        # We can directly copy the data out of a CCITT Group 4 encoded TIFF, if it
        # only contains a single strip
        if (
            img.format == "TIFF"
            and image_filter == "CCITTFaxDecode"
            and img.info["compression"] == "group4"
            and len(img.tag_v2[TiffImagePlugin.STRIPOFFSETS]) == 1  # type: ignore[attr-defined]
            and len(img.tag_v2[TiffImagePlugin.STRIPBYTECOUNTS]) == 1  # type: ignore[attr-defined]
        ):
            photo = img.tag_v2[TiffImagePlugin.PHOTOMETRIC_INTERPRETATION]  # type: ignore[attr-defined]
            inverted = False
            if photo == 0:
                inverted = True
            elif photo != 1:
                raise ValueError(
                    f"unsupported photometric interpretation for g4 tiff: {photo}"
                )
            offset, length = ccitt_payload_location_from_pil(img)
            img_raw_data.seek(offset)
            ccittrawdata: bytes | bytearray = img_raw_data.read(length)
            fillorder = img.tag_v2.get(TiffImagePlugin.FILLORDER)  # type: ignore[attr-defined]
            if fillorder is None or fillorder == 1:
                # no FillOrder or msb-to-lsb: nothing to do
                pass
            elif fillorder == 2:
                # lsb-to-msb: reverse bits of each byte
                ccittrawdata = bytearray(ccittrawdata)
                for i, n in enumerate(ccittrawdata):
                    ccittrawdata[i] = TIFFBitRevTable[n]
                ccittrawdata = bytes(ccittrawdata)
            else:
                raise ValueError(f"unsupported FillOrder: {fillorder}")
            dpn, bpc, colspace = 1, 1, "DeviceGray"
            info.update(
                {
                    "data": ccittrawdata,
                    "w": w,
                    "h": h,
                    "iccp": None,
                    "dpn": dpn,
                    "cs": colspace,
                    "bpc": bpc,
                    "f": image_filter,
                    "inverted": jpeg_inverted,
                    "dp": f"/BlackIs1 {str(not inverted).lower()} /Columns {w} /K -1 /Rows {h}",
                }
            )
            return info

    # garbage collection
    img_raw_data = None

    if img.mode == "1":
        dpn, bpc, colspace = 1, 1, "DeviceGray"
        info["data"] = _to_data(img, image_filter)
    elif img.mode == "L":
        dpn, bpc, colspace = 1, 8, "DeviceGray"
        info["data"] = _to_data(img, image_filter)
    elif img.mode == "LA":
        dpn, bpc, colspace = 1, 8, "DeviceGray"
        alpha_channel = slice(1, None, 2)
        info["data"] = _to_data(img, image_filter, remove_slice=alpha_channel)
        if _has_alpha(img) and image_filter not in (
            "DCTDecode",
            "JPXDecode",
        ):
            info["smask"] = _to_data(img, image_filter, select_slice=alpha_channel)
    elif img.mode == "P":
        dpn, bpc, colspace = 1, 8, "Indexed"
        info["data"] = _to_data(img, image_filter)
        info["pal"] = img.palette.palette if img.palette is not None else None

        # check if the P image has transparency
        if img.info.get("transparency", None) is not None and image_filter not in (
            "DCTDecode",
            "JPXDecode",
        ):
            # convert to RGBA to get the alpha channel for creating the smask
            info["smask"] = _to_data(
                img.convert("RGBA"), image_filter, select_slice=slice(3, None, 4)
            )
    elif img.mode == "PA":
        dpn, bpc, colspace = 1, 8, "Indexed"
        info["pal"] = img.palette.palette if img.palette is not None else None
        alpha_channel = slice(1, None, 2)
        info["data"] = _to_data(img, image_filter, remove_slice=alpha_channel)
        if _has_alpha(img) and image_filter not in (
            "DCTDecode",
            "JPXDecode",
        ):
            info["smask"] = _to_data(img, image_filter, select_slice=alpha_channel)
    elif img.mode == "CMYK":
        dpn, bpc, colspace = 4, 8, "DeviceCMYK"
        info["data"] = _to_data(img, image_filter)
    elif img.mode == "RGB":
        dpn, bpc, colspace = 3, 8, "DeviceRGB"
        info["data"] = _to_data(img, image_filter)
    else:  # RGBA image
        dpn, bpc, colspace = 3, 8, "DeviceRGB"
        alpha_channel = slice(3, None, 4)
        info["data"] = _to_data(img, image_filter, remove_slice=alpha_channel)
        if _has_alpha(img) and image_filter not in (
            "DCTDecode",
            "JPXDecode",
        ):
            info["smask"] = _to_data(img, image_filter, select_slice=alpha_channel)

    dp = f"/Predictor 15 /Colors {dpn} /Columns {w}"

    if img.mode == "1":
        dp = f"/BlackIs1 true /Columns {w} /K -1 /Rows {h}"

    if not is_pil_img:
        if keep_bytes_io_open:
            setattr(img, "fp", None)
        else:
            img.close()

    info.update(
        {
            "w": w,
            "h": h,
            "cs": colspace,
            "iccp": iccp,
            "bpc": bpc,
            "dpn": dpn,
            "f": image_filter,
            "inverted": jpeg_inverted,
            "dp": dp,
        }
    )
    return info


class temp_attr:
    """
    temporary change the attribute of an object using a context manager
    """

    def __init__(self, obj: Any, field: str, value: Any):
        self.obj = obj
        self.field = field
        self.value = value
        self.exists = False
        self.old_value: Any = None

    def __enter__(self) -> None:
        if hasattr(self.obj, self.field):
            self.exists = True
            self.old_value = getattr(self.obj, self.field)
        setattr(self.obj, self.field, self.value)

    def __exit__(self, exctype: Any, excinst: Any, exctb: Any) -> None:
        if self.exists:
            setattr(self.obj, self.field, self.old_value)
        else:
            delattr(self.obj, self.field)


def ccitt_payload_location_from_pil(img: "PILImage") -> tuple[int, int]:
    """
    returns the byte offset and length of the CCITT payload in the original TIFF data
    """
    # assert(img.info["compression"] == "group4")

    # Read the TIFF tags to find the offset(s) of the compressed data strips.
    strip_offsets = img.tag_v2[TiffImagePlugin.STRIPOFFSETS]  # type: ignore[attr-defined]
    strip_bytes = img.tag_v2[TiffImagePlugin.STRIPBYTECOUNTS]  # type: ignore[attr-defined]

    # PIL always seems to create a single strip even for very large TIFFs when
    # it saves images, so assume we only have to read a single strip.
    # A test ~10 GPixel image was still encoded as a single strip. Just to be
    # safe check throw an error if there is more than one offset.
    if (
        len(strip_offsets) != 1  # pyright: ignore[reportUnknownArgumentType]
        or len(strip_bytes) != 1  # pyright: ignore[reportUnknownArgumentType]
    ):  # pyright: ignore[reportUnknownArgumentType]
        raise NotImplementedError(
            "Transcoding multiple strips not supported by the PDF format"
        )

    (offset,), (length,) = (  # pyright: ignore[reportUnknownVariableType]
        strip_offsets,
        strip_bytes,
    )  # pyright: ignore[reportUnknownVariableType]

    return offset, length  # pyright: ignore[reportUnknownVariableType]


def transcode_monochrome(img: "PILImage") -> bytes:
    """
    Convert the open PIL.Image imgdata to compressed CCITT Group4 data.

    """
    # Convert the image to Group 4 in memory. If libtiff is not installed and
    # Pillow is not compiled against it, .save() will raise an exception.
    newimgio = BytesIO()

    # we create a whole new PIL image or otherwise it might happen with some
    # input images, that libtiff fails an assert and the whole process is
    # killed by a SIGABRT:
    img2 = Image.frombytes(img.mode, img.size, img.tobytes())

    # Since version 8.3.0 Pillow limits strips to 64 KB. Since PDF only
    # supports single strip CCITT Group4 payloads, we have to coerce it back
    # into putting everything into a single strip. Thanks to Andrew Murray for
    # the hack.
    #
    # Since version 8.4.0 Pillow allows us to modify the strip size explicitly
    tmp_strip_size = (img.size[0] + 7) // 8 * img.size[1]
    if hasattr(TiffImagePlugin, "STRIP_SIZE"):
        # we are using Pillow 8.4.0 or later
        with temp_attr(TiffImagePlugin, "STRIP_SIZE", tmp_strip_size):
            img2.save(newimgio, format="TIFF", compression="group4")
    else:
        # only needed for Pillow 8.3.x but works for versions before that as
        # well
        pillow__getitem__ = TiffImagePlugin.ImageFileDirectory_v2.__getitem__

        def __getitem__(self: Any, tag: int) -> object:
            overrides = {
                TiffImagePlugin.ROWSPERSTRIP: img.size[1],
                TiffImagePlugin.STRIPBYTECOUNTS: [tmp_strip_size],
                TiffImagePlugin.STRIPOFFSETS: [0],
            }
            return overrides.get(tag, pillow__getitem__(self, tag))

        with temp_attr(
            TiffImagePlugin.ImageFileDirectory_v2, "__getitem__", __getitem__
        ):
            img2.save(newimgio, format="TIFF", compression="group4")

    # Open new image in memory
    newimgio.seek(0)
    newimg = Image.open(newimgio)

    offset, length = ccitt_payload_location_from_pil(newimg)

    newimgio.seek(offset)
    return newimgio.read(length)


def _to_lzwdata(
    img: "PILImage",
    remove_slice: slice | None = None,
    select_slice: slice | None = None,
) -> bytes:
    data = bytearray(img.tobytes())

    if remove_slice:
        del data[remove_slice]
    if select_slice:
        data = data[select_slice]

    if img.mode == "1":
        row_size = ceil(img.size[0] / 8)
    else:
        channels_count = len(data) // (img.size[0] * img.size[1])
        row_size = img.size[0] * channels_count

    data_with_padding = bytearray()
    for i in range(0, len(data), row_size):
        data_with_padding.extend(b"\0")
        data_with_padding.extend(data[i : i + row_size])
    data = data_with_padding
    # Start compression

    # The encoder shall begin by issuing a clear-table code:
    result_codes = [LZW_CLEAR_TABLE_MARKER]
    table, next_code, bits_per_code, max_code_value = clear_table()

    current_sequence = b""
    for byte in data:
        next_sequence = current_sequence + bytes([byte])

        if next_sequence in table:
            # Extend current sequence if already in the table
            current_sequence = next_sequence
        else:
            # Output code for the current sequence
            result_codes.append(table[current_sequence])

            # Add the new sequence to the table if there's room
            if next_code <= (1 << LZW_MAX_BITS_PER_CODE) - 1:
                table[next_sequence] = next_code
                next_code += 1
                if next_code > max_code_value and bits_per_code < LZW_MAX_BITS_PER_CODE:
                    bits_per_code += 1
                    max_code_value = (1 << bits_per_code) - 1
            else:
                # If the table is full, emit a clear-table command
                result_codes.append(LZW_CLEAR_TABLE_MARKER)
                table, next_code, bits_per_code, max_code_value = clear_table()

            # Start new sequence
            current_sequence = bytes([byte])

    # Ensure everything actually is encoded
    if current_sequence:
        result_codes.append(table[current_sequence])

    result_codes.append(LZW_EOD_MARKER)

    return pack_codes_into_bytes(result_codes)


def pack_codes_into_bytes(codes: Iterable[int]) -> bytes:
    """
    Convert the list of result codes into a continuous byte stream, with codes packed as per the code bit-width.
    The bit-width starts at 9 bits and expands as needed.

    """

    (
        _,
        next_code,
        bits_per_code,
        max_code_value,
    ) = clear_table()
    buffer = 0
    bits_in_buffer = 0
    output = bytearray()

    if numpy is not None:
        # Using numpy improves the performance significantly there
        # _cf._ https://github.com/py-pdf/fpdf2/issues/1380
        codes = numpy.array(codes, dtype=numpy.uint32)
    for code in codes:
        buffer = (buffer << bits_per_code) | code
        bits_in_buffer += bits_per_code

        while bits_in_buffer >= 8:
            bits_in_buffer -= 8
            output.append((buffer >> bits_in_buffer) & 0xFF)

        if code == LZW_CLEAR_TABLE_MARKER:
            _, next_code, bits_per_code, max_code_value = clear_table()
        elif code != LZW_EOD_MARKER:
            next_code += 1
            if next_code > max_code_value and bits_per_code < LZW_MAX_BITS_PER_CODE:
                bits_per_code += 1
                max_code_value = (1 << bits_per_code) - 1

    if bits_in_buffer > 0:
        output.append((buffer << (8 - bits_in_buffer)) & 0xFF)

    return bytes(output)


def clear_table() -> tuple[dict[bytes, int], int, int, int]:
    """
    Reset the encoding table and coding state to initial conditions.

    """

    table = {bytes([i]): i for i in range(256)}
    next_code = LZW_EOD_MARKER + 1
    bits_per_code = LZW_INITIAL_BITS_PER_CODE
    max_code_value = (1 << bits_per_code) - 1
    return table, next_code, bits_per_code, max_code_value


def _to_data(img: "PILImage", image_filter: ImageFilter, **kwargs: Any) -> bytes:
    if image_filter == "FlateDecode":
        return _to_zdata(img, **kwargs)

    if image_filter == "CCITTFaxDecode":
        return transcode_monochrome(img)

    if image_filter == "LZWDecode":
        return _to_lzwdata(img, **kwargs)

    if img.mode == "LA":
        img = img.convert("L")

    if img.mode == "RGBA":
        img = img.convert("RGB")

    if image_filter == "DCTDecode":
        compressed_bytes = BytesIO()
        img.save(compressed_bytes, format="JPEG")
        return compressed_bytes.getvalue()

    if image_filter == "JPXDecode":
        compressed_bytes = BytesIO()
        img.save(compressed_bytes, format="JPEG2000")
        return compressed_bytes.getvalue()

    raise FPDFException(f'Unsupported image filter: "{image_filter}"')


def _to_zdata(
    img: "PILImage",
    remove_slice: slice | None = None,
    select_slice: slice | None = None,
) -> bytes:
    data = bytearray(img.tobytes())
    if remove_slice:
        del data[remove_slice]
    if select_slice:
        data = data[select_slice]
    # Left-padding every row with a single zero:
    if img.mode == "1":
        row_size = ceil(img.size[0] / 8)
    else:
        channels_count = len(data) // (img.size[0] * img.size[1])
        row_size = img.size[0] * channels_count

    data_with_padding = bytearray()
    for i in range(0, len(data), row_size):
        data_with_padding.extend(b"\0")
        data_with_padding.extend(data[i : i + row_size])

    return zlib.compress(data_with_padding, level=SETTINGS.compression_level)


def _has_alpha(img: "PILImage") -> bool:
    """
    Return True if the image has any non-opaque alpha channel values.

    For the alpha band, `getextrema()` yields the min & max values across all pixels:
    - (255, 255): every pixel is fully opaque, so we can fast-return False.
    - (x, x) where x != 255: the channel is a flat non-opaque mask, so True.
    Otherwise we need to scan for any value different from 255 (using NumPy if available).
    """
    alpha_channel = img.getchannel("A")
    lo, hi = alpha_channel.getextrema()
    if lo == 255 and hi == 255:
        return False
    if lo == hi:
        return True
    if numpy is not None:
        return (numpy.asarray(alpha_channel) != 255).any()  # type: ignore[no-any-return]
    return any(c != 255 for c in alpha_channel.tobytes())



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_optional_content(tmp_path):  # issue 441
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=24)
    with pdf.optional_content(on_print=False):
        pdf.text(20, 30, "Visible on screen only")
    with pdf.optional_content(on_view=False):
        pdf.text(20, 50, "Printed only")
    pdf.text(20, 70, "Always visible")
    assert_pdf_equal(pdf, HERE / "optional_content.pdf", tmp_path)



import binascii
import socket
import ssl
import urllib.error
from glob import glob
from io import BytesIO
from pathlib import Path
from unittest.mock import patch
from urllib.request import HTTPRedirectHandler, Request

import pytest

import fpdf
from fpdf.image_datastructures import ImageCache

from test.conftest import assert_pdf_equal, ensure_rss_memory_below, time_execution

HERE = Path(__file__).resolve().parent
PNG_BYTES = (HERE / "png_images/c636287a4d7cb1a36362f7f236564cef.png").read_bytes()
SIMPLE_SVG_WITH_REMOTE_IMAGE = b"""
<svg xmlns="http://www.w3.org/2000/svg" width="10" height="10">
  <image href="https://public.example/image.png" x="0" y="0" width="10" height="10"/>
</svg>
"""


class MockResponse(BytesIO):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class MockOpener:
    def __init__(
        self,
        data: bytes,
        redirect_handler=None,
        redirect_to: str | None = None,
    ) -> None:
        self.data = data
        self.redirect_handler = redirect_handler
        self.redirect_to = redirect_to
        self.last_timeout = None

    def open(self, url: str, timeout=None):
        self.last_timeout = timeout
        if self.redirect_to:
            assert self.redirect_handler is not None
            self.redirect_handler.redirect_request(
                Request(url), None, 302, "Found", {}, self.redirect_to
            )
        return MockResponse(self.data)


def _mock_addrinfo_for(*addresses: str):
    addrinfo = []
    for address in addresses:
        family = socket.AF_INET6 if ":" in address else socket.AF_INET
        sockaddr = (address, 0, 0, 0) if family == socket.AF_INET6 else (address, 0)
        addrinfo.append((family, socket.SOCK_STREAM, socket.IPPROTO_TCP, "", sockaddr))
    return addrinfo


def _redirect_handler_from(handlers):
    return next(
        handler for handler in handlers if isinstance(handler, HTTPRedirectHandler)
    )


def test_load_text_file():
    file = HERE / "__init__.py"
    contents = '"""This package contains image tests"""\n'
    bc = contents.encode()

    resource = fpdf.image_parsing.load_image(str(file)).getvalue()
    # loaded a text file in binary mode, may contain DOS style line endings.
    resource = resource.replace(b"\r\n", b"\n")
    assert bytes(resource) == bc


def test_load_base64_data(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(
        "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQBAMAAADt3eJSAAAAMFBMVEU0OkArMjhobHEoPUPFEBIuO0L+AAC2FBZ2JyuNICOfGx7xAwT"
        "jCAlCNTvVDA1aLzQ3COjMAAAAVUlEQVQI12NgwAaCDSA0888GCItjn0szWGBJTVoGSCjWs8TleQCQYV95evdxkFT8Kpe0PLDi5WfKd4LUsN5zS1sKFolt8bwAZrCa"
        "GqNYJAgFDEpQAAAzmxafI4vZWwAAAABJRU5ErkJggg=="
    )
    assert_pdf_equal(pdf, HERE / "load_base64_data.pdf", tmp_path)


def test_load_invalid_base64_data():
    pdf = fpdf.FPDF()
    pdf.add_page()
    with pytest.raises(binascii.Error):
        pdf.image("data:image/png;base64,GARBAGE")


# ensure memory usage does not get too high - this value depends on Python version:
@ensure_rss_memory_below(mib=6)
def test_share_images_cache(tmp_path):
    image_cache = None

    def build_pdf_with_big_images():
        nonlocal image_cache
        pdf = fpdf.FPDF()
        if image_cache is None:
            image_cache = pdf.image_cache
        else:
            pdf.image_cache = image_cache
        pdf.add_page()
        for img_path in glob(f"{HERE}/png_images/*.png"):
            pdf.image(img_path, h=pdf.eph)
        with (tmp_path / "out.pdf").open("wb") as pdf_file:
            pdf.output(pdf_file)
        # Reset the "usages" count:
        image_cache.reset_usages()

    with time_execution() as duration:
        build_pdf_with_big_images()
    first_time_duration = duration.seconds

    with time_execution() as duration:
        build_pdf_with_big_images()
    assert duration.seconds < first_time_duration / 2


def test_load_image_blocks_local_files_by_policy():
    file = HERE / "__init__.py"
    with pytest.raises(fpdf.FPDFResourceAccessError):
        fpdf.image_parsing.load_image(
            str(file), resource_access_policy=fpdf.ResourceAccessPolicy.NONE
        )


def test_load_image_allows_public_remote_by_default():
    opener = MockOpener(PNG_BYTES)
    with (
        patch(
            "fpdf.image_parsing.socket.getaddrinfo",
            return_value=_mock_addrinfo_for("93.184.216.34"),
        ),
        patch(
            "fpdf.image_parsing.build_opener",
            side_effect=lambda *handlers: setattr(
                opener, "redirect_handler", _redirect_handler_from(handlers)
            )
            or opener,
        ),
    ):
        resource = fpdf.image_parsing.load_image("https://public.example/image.png")
    assert resource.getvalue() == PNG_BYTES
    assert opener.last_timeout == fpdf.image_parsing.SETTINGS.network_timeout


def test_load_image_blocks_private_remote_by_default():
    with patch(
        "fpdf.image_parsing.socket.getaddrinfo",
        return_value=_mock_addrinfo_for("127.0.0.1"),
    ):
        with pytest.raises(fpdf.FPDFResourceAccessError):
            fpdf.image_parsing.load_image("http://private.example/image.png")


def test_load_image_blocks_private_redirect_by_default():
    def fake_getaddrinfo(host, *_args, **_kwargs):
        if host == "public.example":
            return _mock_addrinfo_for("93.184.216.34")
        if host == "private.example":
            return _mock_addrinfo_for("127.0.0.1")
        raise AssertionError(f"Unexpected hostname: {host}")

    with patch("fpdf.image_parsing.socket.getaddrinfo", side_effect=fake_getaddrinfo):
        with patch(
            "fpdf.image_parsing.build_opener",
            side_effect=lambda *handlers: MockOpener(
                PNG_BYTES,
                redirect_handler=_redirect_handler_from(handlers),
                redirect_to="http://private.example/image.png",
            ),
        ):
            with pytest.raises(fpdf.FPDFResourceAccessError):
                fpdf.image_parsing.load_image("https://public.example/image.png")


def test_load_image_ignores_dns_rebinding_after_validation():
    resolution_count = {"count": 0}
    create_connection_calls = []

    def fake_getaddrinfo(host, *_args, **_kwargs):
        assert host == "attacker.example"
        resolution_count["count"] += 1
        if resolution_count["count"] == 1:
            return _mock_addrinfo_for("93.184.216.34")
        return _mock_addrinfo_for("127.0.0.1")

    def fake_create_connection(address, *_args, **_kwargs):
        create_connection_calls.append(address)
        raise OSError("stop before network")

    with (
        patch("fpdf.image_parsing.socket.getaddrinfo", side_effect=fake_getaddrinfo),
        patch(
            "fpdf.image_parsing.socket.create_connection",
            side_effect=fake_create_connection,
        ),
    ):
        with pytest.raises(urllib.error.URLError):
            fpdf.image_parsing.load_image(
                "http://attacker.example/image.png",
                resource_access_policy=fpdf.ResourceAccessPolicy.REMOTE_PUBLIC,
            )

    assert resolution_count["count"] == 1
    assert create_connection_calls == [("93.184.216.34", 80)]


def test_load_image_pins_validated_address_for_connection():
    create_connection_calls = []

    def fake_create_connection(address, *_args, **_kwargs):
        create_connection_calls.append(address)
        raise OSError("stop before network")

    with (
        patch(
            "fpdf.image_parsing.socket.getaddrinfo",
            return_value=_mock_addrinfo_for("93.184.216.34"),
        ),
        patch(
            "fpdf.image_parsing.socket.create_connection",
            side_effect=fake_create_connection,
        ),
    ):
        with pytest.raises(urllib.error.URLError):
            fpdf.image_parsing.load_image(
                "http://public.example/image.png",
                resource_access_policy=fpdf.ResourceAccessPolicy.REMOTE_PUBLIC,
            )

    assert create_connection_calls == [("93.184.216.34", 80)]


def test_load_image_https_pinning_preserves_sni_hostname():
    create_connection_calls = []
    server_hostnames = []

    class FakeSocket:
        @staticmethod
        def setsockopt(*_args, **_kwargs):
            return None

        @staticmethod
        def close():
            return None

    def fake_create_connection(address, *_args, **_kwargs):
        create_connection_calls.append(address)
        return FakeSocket()

    def fake_wrap_socket(_self, _sock, *_args, server_hostname=None, **_kwargs):
        server_hostnames.append(server_hostname)
        raise OSError("stop before network")

    with (
        patch(
            "fpdf.image_parsing.socket.getaddrinfo",
            return_value=_mock_addrinfo_for("93.184.216.34"),
        ),
        patch(
            "fpdf.image_parsing.socket.create_connection",
            side_effect=fake_create_connection,
        ),
        patch.object(ssl.SSLContext, "wrap_socket", fake_wrap_socket),
    ):
        with pytest.raises(urllib.error.URLError):
            fpdf.image_parsing.load_image(
                "https://public.example/image.png",
                resource_access_policy=fpdf.ResourceAccessPolicy.REMOTE_PUBLIC,
            )

    assert create_connection_calls == [("93.184.216.34", 443)]
    assert server_hostnames == ["public.example"]


def test_svg_preload_preserves_resource_access_error_for_blocked_redirect():
    def fake_getaddrinfo(host, *_args, **_kwargs):
        if host == "public.example":
            return _mock_addrinfo_for("93.184.216.34")
        if host == "private.example":
            return _mock_addrinfo_for("127.0.0.1")
        raise AssertionError(f"Unexpected hostname: {host}")

    with (
        patch("fpdf.image_parsing.socket.getaddrinfo", side_effect=fake_getaddrinfo),
        patch(
            "fpdf.image_parsing.build_opener",
            side_effect=lambda *handlers: MockOpener(
                b'<svg xmlns="http://www.w3.org/2000/svg" width="10" height="10"/>',
                redirect_handler=_redirect_handler_from(handlers),
                redirect_to="http://private.example/image.svg",
            ),
        ),
    ):
        with pytest.raises(fpdf.FPDFResourceAccessError):
            fpdf.image_parsing.preload_image(
                ImageCache(), "https://public.example/image.svg"
            )


def test_image_per_call_policy_override_allows_private_remote():
    pdf = fpdf.FPDF()
    pdf.resource_access_policy = fpdf.ResourceAccessPolicy.NONE
    pdf.add_page()
    with (
        patch(
            "fpdf.image_parsing.socket.getaddrinfo",
            return_value=_mock_addrinfo_for("127.0.0.1"),
        ),
        patch(
            "fpdf.image_parsing.build_opener",
            side_effect=lambda *handlers: MockOpener(
                PNG_BYTES, redirect_handler=_redirect_handler_from(handlers)
            ),
        ),
    ):
        pdf.image(
            "http://private.example/image.png",
            x=10,
            y=10,
            w=10,
            resource_access_policy=fpdf.ResourceAccessPolicy.ALL,
        )


def test_svg_image_per_call_policy_override_applies_to_nested_resources(tmp_path):
    svg_path = tmp_path / "nested.svg"
    svg_path.write_text(
        """
        <svg xmlns="http://www.w3.org/2000/svg" width="10" height="10">
        <image href="http://private.example/image.png" x="0" y="0" width="10" height="10"/>
        </svg>""",
        encoding="utf-8",
    )

    pdf = fpdf.FPDF()
    pdf.resource_access_policy = fpdf.ResourceAccessPolicy.NONE
    pdf.add_page()
    with (
        patch(
            "fpdf.image_parsing.socket.getaddrinfo",
            return_value=_mock_addrinfo_for("127.0.0.1"),
        ),
        patch(
            "fpdf.image_parsing.build_opener",
            side_effect=lambda *handlers: MockOpener(
                PNG_BYTES, redirect_handler=_redirect_handler_from(handlers)
            ),
        ),
    ):
        pdf.image(
            svg_path,
            w=10,
            resource_access_policy=fpdf.ResourceAccessPolicy.ALL,
        )


def test_svg_object_draw_to_page_uses_document_resource_access_policy():
    svg = fpdf.svg.SVGObject(SIMPLE_SVG_WITH_REMOTE_IMAGE)
    pdf = fpdf.FPDF()
    pdf.resource_access_policy = fpdf.ResourceAccessPolicy.NONE
    pdf.add_page()

    with (
        patch(
            "fpdf.image_parsing.socket.getaddrinfo",
            return_value=_mock_addrinfo_for("93.184.216.34"),
        ),
        patch(
            "fpdf.image_parsing.build_opener",
            side_effect=lambda *handlers: MockOpener(
                PNG_BYTES, redirect_handler=_redirect_handler_from(handlers)
            ),
        ),
    ):
        with pytest.raises(fpdf.FPDFResourceAccessError):
            svg.draw_to_page(pdf)

    assert svg.resource_access_policy == fpdf.ResourceAccessPolicy.DEFAULT


def test_write_html_uses_document_resource_access_policy():
    img_path = HERE / "png_images/c636287a4d7cb1a36362f7f236564cef.png"
    pdf = fpdf.FPDF()
    pdf.resource_access_policy = fpdf.ResourceAccessPolicy.NONE
    pdf.add_page()
    with pytest.raises(fpdf.FPDFResourceAccessError):
        pdf.write_html(f'<img src="{img_path}" width="10" height="10">')


def test_template_image_uses_document_resource_access_policy():
    img_path = HERE / "png_images/c636287a4d7cb1a36362f7f236564cef.png"
    pdf = fpdf.FPDF()
    pdf.resource_access_policy = fpdf.ResourceAccessPolicy.NONE
    pdf.add_page()
    template = fpdf.FlexTemplate(
        pdf,
        elements=[
            {
                "name": "img",
                "type": "I",
                "x1": 10,
                "y1": 10,
                "x2": 20,
                "y2": 20,
                "text": str(img_path),
            }
        ],
    )
    with pytest.raises(fpdf.FPDFResourceAccessError):
        template.render()


def test_text_columns_downscale_reload_uses_document_resource_access_policy():
    def fake_getaddrinfo(host, *_args, **_kwargs):
        if host == "public.example":
            return _mock_addrinfo_for("93.184.216.34")
        if host == "private.example":
            return _mock_addrinfo_for("127.0.0.1")
        raise AssertionError(f"Unexpected hostname: {host}")

    open_count = {"count": 0}

    class RedirectOnSecondOpen:
        def __init__(self, *handlers) -> None:
            self.redirect_handler = _redirect_handler_from(handlers)

        def open(self, url: str, timeout=None):  # pylint: disable=unused-argument
            open_count["count"] += 1
            if open_count["count"] == 2:
                self.redirect_handler.redirect_request(
                    Request(url),
                    None,
                    302,
                    "Found",
                    {},
                    "http://private.example/image.png",
                )
            return MockResponse(PNG_BYTES)

    pdf = fpdf.FPDF()
    pdf.resource_access_policy = fpdf.ResourceAccessPolicy.REMOTE_PUBLIC
    pdf.oversized_images = "downscale"
    pdf.add_page()

    with (
        patch("fpdf.image_parsing.socket.getaddrinfo", side_effect=fake_getaddrinfo),
        patch(
            "fpdf.image_parsing.build_opener",
            side_effect=RedirectOnSecondOpen,
        ),
    ):
        with pytest.raises(fpdf.FPDFResourceAccessError):
            with pdf.text_columns() as cols:
                cols.image("https://public.example/image.png", width=1, height=1)



from pathlib import Path

import pytest

import fpdf
from fpdf.errors import FPDFSvgLimitExceeded
from fpdf.svg import SVGObject, SVGLimits


def _nested_use_svg(level: int, fanout: int = 2) -> bytes:
    groups = ['<g id="g0"><rect width="1" height="1"/></g>']
    for current_level in range(1, level + 1):
        uses = f'<use href="#g{current_level - 1}"/>' * fanout
        groups.append(f'<g id="g{current_level}">{uses}</g>')
    return (
        '<svg xmlns="http://www.w3.org/2000/svg" width="1" height="1">'
        "<defs>" + "".join(groups) + f'</defs><use href="#g{level}"/></svg>'
    ).encode()


def _reverse_nested_use_svg(level: int) -> bytes:
    groups = [
        f'<g id="g{current_level}"><use href="#g{current_level - 1}"/></g>'
        for current_level in range(level, 0, -1)
    ]
    groups.append('<g id="g0"><rect width="1" height="1"/></g>')
    return (
        '<svg xmlns="http://www.w3.org/2000/svg" width="1" height="1">'
        "<defs>" + "".join(groups) + f'</defs><use href="#g{level}"/></svg>'
    ).encode()


def test_default_svg_limits() -> None:
    pdf = fpdf.FPDF()

    assert isinstance(pdf.svg_limits, SVGLimits)
    assert pdf.svg_limits == SVGLimits()


def test_fpdf_svg_limits_can_be_replaced() -> None:
    pdf = fpdf.FPDF()
    custom_limits = SVGLimits(
        max_use_depth=64,
        max_resolved_elements=500_000,
    )

    pdf.svg_limits = custom_limits

    assert pdf.svg_limits is custom_limits


@pytest.mark.parametrize("max_use_depth", [0, -1])
def test_svg_limits_reject_non_positive_use_depth(max_use_depth: int) -> None:
    with pytest.raises(ValueError, match="must be a positive integer or None"):
        SVGLimits(max_use_depth=max_use_depth)


@pytest.mark.parametrize("max_resolved_elements", [0, -1])
def test_svg_limits_reject_non_positive_resolved_elements(
    max_resolved_elements: int,
) -> None:
    with pytest.raises(ValueError, match="must be a positive integer or None"):
        SVGLimits(max_resolved_elements=max_resolved_elements)


def test_svg_limits_reject_use_cycles() -> None:
    svg = b"""
    <svg xmlns="http://www.w3.org/2000/svg" width="1" height="1">
      <defs>
        <g id="a"><use href="#b"/></g>
        <g id="b"><use href="#a"/></g>
      </defs>
      <use href="#a"/>
    </svg>
    """

    with pytest.raises(fpdf.FPDFException, match="cycle"):
        SVGObject(svg)


def test_svg_limits_reject_excessive_use_depth() -> None:
    with pytest.raises(FPDFSvgLimitExceeded, match="max_use_depth"):
        SVGObject(_nested_use_svg(level=2), svg_limits=SVGLimits(max_use_depth=1))


def test_svg_limits_reject_deep_reverse_ordered_use_chain() -> None:
    with pytest.raises(FPDFSvgLimitExceeded, match="max_use_depth"):
        SVGObject(_reverse_nested_use_svg(level=1100))


def test_image_uses_svg_resolved_element_limit() -> None:
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.svg_limits = SVGLimits(max_resolved_elements=20)

    with pytest.raises(fpdf.FPDFException, match="max_resolved_elements"):
        pdf.image(_nested_use_svg(level=2, fanout=4), w=10)


def test_image_preserves_limit_exception(tmp_path: Path) -> None:
    svg_file = tmp_path / "heavy.svg"
    svg_file.write_bytes(_nested_use_svg(level=2, fanout=4))
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.svg_limits = SVGLimits(max_resolved_elements=20)

    with pytest.raises(fpdf.FPDFException, match="max_resolved_elements"):
        pdf.image(svg_file, w=10)


def test_svg_limits_count_switch_elements() -> None:
    svg = b"""
    <svg xmlns="http://www.w3.org/2000/svg" width="1" height="1">
      <switch>
        <rect width="1" height="1"/>
      </switch>
    </svg>
    """

    with pytest.raises(FPDFSvgLimitExceeded, match="max_resolved_elements"):
        SVGObject(svg, svg_limits=SVGLimits(max_resolved_elements=2))


def test_svg_limits_ignore_skipped_switch_children() -> None:
    svg = b"""
    <svg xmlns="http://www.w3.org/2000/svg" width="1" height="1">
      <switch>
        <g requiredFeatures="http://example.invalid/feature">
          <rect width="1" height="1"/>
          <rect width="1" height="1"/>
          <rect width="1" height="1"/>
        </g>
        <circle cx="0.5" cy="0.5" r="0.5"/>
      </switch>
    </svg>
    """

    assert SVGObject(svg, svg_limits=SVGLimits(max_resolved_elements=3))


def test_disable_complexity_limits() -> None:
    svg = SVGObject(
        _nested_use_svg(level=3, fanout=4),
        svg_limits=SVGLimits(max_use_depth=None, max_resolved_elements=None),
    )

    assert svg.base_group is not None



import itertools
from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal
from test.conftest import LOREM_IPSUM

import pytest

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_multi_cell_markdown(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum dolor amet, **consectetur adipiscing** elit,"
        " sed do eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown.pdf", tmp_path)


def test_multi_cell_markdown_strikethrough(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=32)
    pdf.multi_cell(w=pdf.epw, text="~~strikethrough~~", markdown=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_strikethrough.pdf", tmp_path)


def test_multi_cell_markdown_escaped(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum \\ dolor amet, \\**consectetur adipiscing\\** elit,"
        " sed do eiusmod \\\\__tempor incididunt\\\\__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_escaped.pdf", tmp_path)


def test_multi_cell_markdown_with_ttf_fonts(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum dolor, **consectetur adipiscing** elit,"
        " eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_with_ttf_fonts.pdf", tmp_path)


def test_multi_cell_markdown_with_ttf_fonts_escaped(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum \\ dolor, \\**consectetur adipiscing\\** elit,"
        " eiusmod \\\\__tempor incididunt\\\\__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(
        pdf, HERE / "multi_cell_markdown_with_ttf_fonts_escaped.pdf", tmp_path
    )


def test_multi_cell_markdown_missing_ttf_font():
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.add_font(fname=FONTS_DIR / "Roboto-Regular.ttf")
    pdf.set_font("Roboto-Regular", size=60)
    with pytest.raises(fpdf.FPDFException) as error:
        pdf.multi_cell(w=pdf.epw, text="**Lorem Ipsum**", markdown=True)
    expected_msg = "Undefined font: roboto-regularB - Use built-in fonts or FPDF.add_font() beforehand"
    assert str(error.value) == expected_msg


def test_multi_cell_markdown_with_fill_color(tmp_path):  # issue 348
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=10)
    pdf.set_fill_color(255, 0, 0)
    pdf.multi_cell(
        50, markdown=True, text="aa bb cc **dd ee dd ee dd ee dd ee dd ee dd ee**"
    )
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_with_fill_color.pdf", tmp_path)


def test_multi_cell_markdown_justified(tmp_path):  # issue 327
    pdf = fpdf.FPDF()
    pdf.add_page()
    for font in ("Helvetica", "Courier"):
        pdf.set_font(family=font, size=12)
        pdf.set_y(pdf.y + 3)
        pdf.multi_cell(
            190,
            markdown=True,
            align="J",
            text=(
                "Lorem **ipsum** dolor sit amet, **consectetur** adipiscing elit, "
                "sed do eiusmod tempor incididunt ut labore et dolore magna "
                "aliqua. Ut enim ad minim veniam, __quis__ nostrud exercitation "
                "ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis "
                "aute irure dolor in reprehenderit in voluptate velit esse cillum "
                "dolore eu fugiat nulla pariatur. Excepteur sint occaecat "
                "cupidatat non proident, sunt in culpa qui officia deserunt "
                "mollit anim id est laborum."
            ),
        )
        pdf.set_x(10)
        pdf.set_y(pdf.y + 5)
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_justified.pdf", tmp_path)


def test_multi_cell_markdown_link(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    pdf.multi_cell(
        pdf.epw,
        text="**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2) __End__",
        markdown=True,
    )
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_link.pdf", tmp_path)


def test_multi_cell_markdown_link_dry_run(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    assert len(pdf.pages[1].annots) == 0

    pdf.multi_cell(
        pdf.epw,
        text="**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2) __End__",
        dry_run=True,
        markdown=True,
        new_x="left",
        new_y="next",
    )
    assert len(pdf.pages[1].annots) == 0

    pdf.multi_cell(
        pdf.epw,
        text="**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2) __End__",
        markdown=True,
        new_x="left",
        new_y="next",
    )
    assert len(pdf.pages[1].annots) == 1

    pdf.multi_cell(
        pdf.epw,
        text="**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2) __End__",
        dry_run=True,
        markdown=True,
        new_x="left",
        new_y="next",
    )
    assert len(pdf.pages[1].annots) == 1

    pdf.multi_cell(
        pdf.epw,
        text="**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2) __End__",
        markdown=True,
        new_x="left",
        new_y="next",
    )
    assert len(pdf.pages[1].annots) == 2

    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_link_dry_run.pdf", tmp_path)


def test_multi_cell_markdown_consecutive_links(tmp_path):
    link1 = "[fpdf2 github](https://github.com/py-pdf/fpdf2)"
    link2 = "[fpdf2 github Releases](https://github.com/py-pdf/fpdf2/releases)"

    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    pdf.multi_cell(
        pdf.epw,
        text=f"**Start** {link1:s} {link2:s} __End__",
        markdown=True,
        new_x="left",
        new_y="next",
    )
    assert len(pdf.pages[pdf.page].annots) == 2
    pdf.multi_cell(
        pdf.epw,
        text=f"**Start** {link1:s}{link2:s} __End__",
        markdown=True,
        new_x="left",
        new_y="next",
    )
    assert len(pdf.pages[pdf.page].annots) == 4
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_consecutive_links.pdf", tmp_path)


def test_multi_cell_markdown_styled_link(tmp_path):
    styles = (
        ("Bold", "**"),
        ("Italics", "__"),
        ("Strikethrough", "~~"),
        ("Underline", "--"),
    )
    style_combinations = []
    for i in range(1, len(styles) + 1):
        for combo in itertools.combinations(styles, i):
            style = "-".join(c[0] for c in combo)
            marker = "".join(c[1] for c in combo)
            style_combinations.append((style, marker))

    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()

    for link_color, link_underline in itertools.product(
        (None, "#0000ff"),
        (False, True),
    ):
        pdf.MARKDOWN_LINK_COLOR = link_color
        pdf.MARKDOWN_LINK_UNDERLINE = link_underline
        for style, marker in style_combinations:
            pdf.multi_cell(
                pdf.epw,
                text=f"**Start** {marker:s}[{style:s} Link](https://github.com/py-pdf/fpdf2){marker:s} __End__",
                markdown=True,
                new_x="left",
                new_y="next",
            )
        pdf.ln()

    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_styled_link.pdf", tmp_path)


@pytest.mark.parametrize(
    "text",
    [
        "**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2)\n__End__",
        LOREM_IPSUM
        + "\n**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2)\n__End__",
        LOREM_IPSUM[: len(LOREM_IPSUM) // 2]
        + " **\nStart** [fpdf2 github](https://github.com/py-pdf/fpdf2)\n__End__ "
        + LOREM_IPSUM[len(LOREM_IPSUM) // 2 :],
    ],
)
def test_multi_cell_markdown_dry_run_lines_output(text):
    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()

    lines = pdf.multi_cell(
        pdf.epw,
        text=text,
        dry_run=True,
        markdown=True,
        new_x="left",
        new_y="next",
        output=fpdf.enums.MethodReturnValue.LINES,
    )

    # The parts of the special markdown text must be in the lines list, but not
    # in the same line
    assert any("**Start**" in line for line in lines)
    assert any(
        "[fpdf2 github](https://github.com/py-pdf/fpdf2)" in line for line in lines
    )
    assert any("__End__" in line for line in lines)
    start_line = next(i for i, line in enumerate(lines) if "**Start**" in line)
    end_line = next(i for i, line in enumerate(lines) if "__End__" in line)
    assert start_line + 1 == end_line

    parsed_text = "\n".join(lines)
    assert (
        "**Start** [fpdf2 github](https://github.com/py-pdf/fpdf2)\n__End__"
        in parsed_text
    )


def test_multi_cell_markdown_dry_run_lines_output_print(tmp_path):
    # Test that output="LINES" keeps markdown format
    text = (
        LOREM_IPSUM[: len(LOREM_IPSUM) // 2]
        + "\n**Start** ~~test~~ "
        + "[fpdf2 github](https://github.com/py-pdf/fpdf2) "
        + "--test--\n__End__ "
        + LOREM_IPSUM[len(LOREM_IPSUM) // 2 :]
    )

    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()

    # Normal text
    pdf.multi_cell(
        pdf.epw,
        text=text,
        markdown=True,
        new_x="left",
        new_y="next",
    )
    pdf.ln()

    # Join text after dry run by `"\n"`
    lines = pdf.multi_cell(
        pdf.epw,
        text=text,
        dry_run=True,
        markdown=True,
        new_x="left",
        new_y="next",
        output=fpdf.enums.MethodReturnValue.LINES,
    )
    pdf.multi_cell(
        pdf.epw,
        text="\n".join(lines),
        markdown=True,
        new_x="left",
        new_y="next",
    )

    assert_pdf_equal(
        pdf, HERE / "multi_cell_markdown_dry_run_lines_output.pdf", tmp_path
    )


def test_multi_cell_markdown_dry_run_lines_output_escape(tmp_path):
    # Test that escaped markdown markers stay escaped
    text = (
        LOREM_IPSUM[: len(LOREM_IPSUM) // 2]
        + "\n**Start** \\** [fpdf2 **github**](https://github.com/py-pdf/fpdf2) "
        + "\\__ \\~~ \\--\n__End__ "  # Important test underline after link
        + LOREM_IPSUM[len(LOREM_IPSUM) // 2 :]
    )

    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()

    # Normal text
    pdf.multi_cell(
        pdf.epw,
        text=text,
        markdown=True,
        new_x="left",
        new_y="next",
    )
    pdf.ln()

    # Join text after dry run by `"\n"`
    lines = pdf.multi_cell(
        pdf.epw,
        text=text,
        dry_run=True,
        markdown=True,
        new_x="left",
        new_y="next",
        output=fpdf.enums.MethodReturnValue.LINES,
    )
    pdf.multi_cell(
        pdf.epw,
        text="\n".join(lines),
        markdown=True,
        new_x="left",
        new_y="next",
    )

    assert_pdf_equal(
        pdf, HERE / "multi_cell_markdown_dry_run_lines_output_escape.pdf", tmp_path
    )



from pathlib import Path

from fpdf import FPDF

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_cid_keyed_cff_font_embeds_cff_font_program(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Shsans", fname=HERE / "SourceHanSansCN-Normal.otf")

    pdf.set_font("Shsans", size=14)
    pdf.multi_cell(
        w=0,
        text=(
            "Lorem ipsum dolor sit amet, consectetur adipisicing elit. Culpa labore "
            "quas fuga magnam, tempora repellendus ipsa, provident officiis adipisci "
            "modi ab perspiciatis ducimus pariatur dolor odio placeat odit atque? "
            "Aperiam. 天地玄黄 宇宙洪荒 日月盈昃 辰宿列张 寒来暑往 秋收冬藏 "
            "闰余成岁 律吕调阳"
        ),
    )

    assert_pdf_equal(
        pdf,
        HERE / "font_opentype.pdf",
        tmp_path,
    )
